import collections
import os
import pathlib
import yaml


InputOnlyCase = collections.namedtuple("InputOnlyCase", ["name", "content"])
InputOutputCase = collections.namedtuple(
    "InputOutputCase", ["name", "given", "expected", "config"]
)


def has_extension(expected_extension):
    def verify(filename):
        extension = "".join(pathlib.Path(filename).suffixes)
        return extension == expected_extension

    return verify


def get_directory_path(directory):
    return os.path.join(os.path.dirname(os.path.realpath(__file__)), directory)


def get_files_with_extension(directory, extension):
    files = os.listdir(get_directory_path(directory))
    return list(filter(has_extension(extension), files))


def remove_extension(filename):
    return filename.replace("".join(pathlib.Path(filename).suffixes), "")


def get_content(filename, directory):
    filepath = os.path.join(get_directory_path(directory), filename)
    with open(filepath, "r", encoding="utf-8") as opened_file:
        return opened_file.read()


def create_tests_generator(cases):
    def tests_generator(metafunc):
        if "case" in metafunc.fixturenames:
            metafunc.parametrize(
                argnames="case", argvalues=cases, ids=[case.name for case in cases]
            )

    return tests_generator


def discover_input_only_cases(where, input_extension):
    return [
        InputOnlyCase(remove_extension(f), get_content(f, directory=where))
        for f in get_files_with_extension(where, input_extension)
    ]


def get_matching_output_filename(input_filename, output_extension):
    return f"{remove_extension(input_filename)}{output_extension}"


def make_input_output_case(input_filename, output_extension, where):
    output_filename = get_matching_output_filename(input_filename, output_extension)
    given = get_content(input_filename, directory=where)
    if given.startswith("###"):
        head, *rest = given.splitlines()
        given = "\n".join(rest)

        config = yaml.safe_load(head[3:])
    else:
        config = dict()

    return InputOutputCase(
        remove_extension(input_filename),
        given,
        get_content(output_filename, directory=where),
        config,
    )


def discover_input_output_cases(where, input_extension, output_extension):
    input_files = get_files_with_extension(where, input_extension)
    output_files = get_files_with_extension(where, output_extension)
    for inp in input_files:
        matching_output_file = get_matching_output_filename(inp, output_extension)
        assert (
            matching_output_file in output_files
        ), f"Incomplete input-output pair, missing {matching_output_file}"

    return [make_input_output_case(inp, output_extension, where) for inp in input_files]


def generate_input_only_tests(where, input_extension):
    return create_tests_generator(discover_input_only_cases(where, input_extension))


def generate_input_output_tests(where, input_extension, output_extension):
    return create_tests_generator(
        discover_input_output_cases(where, input_extension, output_extension)
    )



from gersemi.ast_helpers import is_line_comment, is_commented_argument
from gersemi.base_dumper import BaseDumper


class CommandLineFormatter(BaseDumper):
    def _should_start_new_line(self, updated_line, last_line):
        return (
            len(updated_line) > self.width
            or "\n" in updated_line
            or last_line.strip().startswith("#")
        )

    def _format_command_line(self, args):
        head, *tail = args
        lines = [self.visit(head)]
        force_next_line = is_commented_argument(head)
        for arg in tail:
            if is_line_comment(arg):
                lines += [self.visit(arg)]
                continue

            with self.not_indented():
                formatted_arg = self.visit(arg)
            updated_line = f"{lines[-1]} {formatted_arg}"
            if force_next_line or self._should_start_new_line(updated_line, lines[-1]):
                force_next_line = is_commented_argument(arg)
                lines += [self.visit(arg)]
            else:
                lines[-1] = updated_line
                if is_commented_argument(arg):
                    force_next_line = True

        return "\n".join(lines)



from enum import Enum
import textwrap


def doc(text: str) -> str:
    return " ".join(textwrap.dedent(text).splitlines()).strip()


class EnumWithMetadata(Enum):
    def __new__(cls, data):
        self = object.__new__(cls)
        self._value_ = data["value"]
        self.description = doc(data["description"])
        self.title = data["title"]
        return self

    def __repr__(self):
        return self.value



from dataclasses import asdict, fields
import os
from pathlib import Path
from typing import Iterable, Optional
from gersemi.configuration import (
    make_configuration_file,
    OutcomeConfiguration,
)
from gersemi.extensions import load_definitions_from_extension


NL = "\n"


def report_header(configuration_file: Optional[Path]) -> str:
    if configuration_file is None:
        based_on = "defaults"
    else:
        based_on = f"{configuration_file.resolve()}"

    return f"## Outcome configuration based on {based_on}"


def prepare_configuration_for_report(
    configuration_file: Optional[Path], configuration: OutcomeConfiguration
):
    result = asdict(configuration)
    if configuration_file is not None:
        result["definitions"] = tuple(
            os.path.relpath(p, configuration_file.parent)
            for p in result.get("definitions", tuple())
        )

    return result


def minimal_report(
    configuration_file: Optional[Path], configuration: OutcomeConfiguration
) -> str:
    c = prepare_configuration_for_report(configuration_file, configuration)
    filtered = {}
    keys = {f.name: f for f in fields(OutcomeConfiguration)}
    for name, value in c.items():
        if value != keys[name].default:
            filtered[name] = value

    if filtered == dict():
        comparison = "none of the defaults are overridden.\n"
    else:
        comparison = f"""following options differ from default configuration:
{make_configuration_file(filtered, add_schema_link=False)}"""

    return f"{report_header(configuration_file)},\n## {comparison}\n"


def make_listing(prefix, things):
    return f"{prefix}{f'{NL}{prefix}'.join(things)}"


def applicable_to_files(files: Iterable[Path]) -> str:
    if Path("-") in files:
        return "## it's applicable to stdin."

    return f"""## it's applicable to these files:
{make_listing("## - ", map(str, files))}
##"""


def verbose_about_extension(extension: str) -> str:
    definitions, error = load_definitions_from_extension(extension)
    if error is None:
        return f"""## - [{extension}] additional recognized commands:
{make_listing("##   - ", sorted(definitions.keys()))}
##"""

    return f"""## - [{extension}] error:
{make_listing("##   ", error.splitlines())}
##"""


def verbose_about_extensions(extensions) -> str:
    if len(extensions) == 0:
        return ""

    return f"""
## About extensions:
{NL.join(map(verbose_about_extension, extensions))}"""


def verbose_report(
    configuration_file: Optional[Path],
    configuration: OutcomeConfiguration,
    target_files: Iterable[Path],
) -> str:
    listing = make_configuration_file(
        prepare_configuration_for_report(configuration_file, configuration),
        add_schema_link=True,
    )

    return f"""{report_header(configuration_file)},
{applicable_to_files(target_files)}{verbose_about_extensions(configuration.extensions)}
{listing}
"""


def default_report():
    return make_configuration_file(
        asdict(OutcomeConfiguration()),
        add_schema_link=True,
    )



# pylint: disable=redefined-outer-name
from functools import partial
import pytest
from tests.fixtures.app import success


@pytest.fixture(scope="function")
def base_directory(testfiles):
    return testfiles / "extensions" / "example"


@pytest.fixture(scope="function")
def app(app):
    return partial(app, "--extensions", "extension_example")


@pytest.mark.parametrize("case", ["favour-inlining", "favour-expansion"])
def test_extension_example(app, base_directory, case):
    file_to_format = base_directory / "code_to_format" / f"{case}.cmake"
    with open(file_to_format, "r", encoding="utf-8") as f:
        code = f.read()

    assert app("--list-expansion", case, "--", file_to_format) == success(stdout=code)
    assert app("--list-expansion", case, "--check", file_to_format) == success()



import collections.abc
from contextlib import contextmanager
from functools import lru_cache
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.specializations.preserving_command_invocation_dumper import (
    PreservingCommandInvocationDumper,
)
from gersemi.specializations.standard_command_dumper import (
    create_standard_dumper,
)


@lru_cache(maxsize=None)
def create_patch(patch, old_class):
    class Impl(patch, old_class):
        pass

    return Impl


class CommandInvocationDumper(
    PreservingCommandInvocationDumper, BaseCommandInvocationDumper
):
    @contextmanager
    def patched(self, patch):
        old_class = type(self)
        try:
            # pylint: disable=attribute-defined-outside-init
            self.__class__ = create_patch(patch, old_class)
            yield self
        finally:
            self.__class__ = old_class  # pylint: disable=invalid-class-object,

    def _get_patch(self, raw_command_name):
        command = self.known_definitions.get(raw_command_name.lower(), None)
        if isinstance(command, collections.abc.Mapping):
            return create_standard_dumper(command)

        return command

    def command_invocation(self, tree):
        command_name, _ = tree.children
        patch = self._get_patch(command_name)
        if patch is None:
            return super().format_command(tree)
        with self.patched(patch):
            return self.format_command(tree)

    def custom_command(self, tree):
        _, command_name, *_ = tree.children
        self._record_unknown_command(command_name)
        return super().custom_command(tree)



# pylint: disable=too-many-lines
from contextlib import contextmanager, ExitStack
import filecmp
from functools import partial
import os
from pathlib import Path
import sqlite3
from stat import S_IREAD, S_IRGRP, S_IROTH
import yaml
import pytest
from gersemi.return_codes import FAIL, SUCCESS
from tests.fixtures.app import ExpectedOutcome, fail, match_not, success


def compare_directories(left, right):
    comparison = filecmp.dircmp(left, right)
    return {
        "left_only": comparison.left_only,
        "right_only": comparison.right_only,
        "diff_files": comparison.diff_files,
        "funny_files": comparison.funny_files,
    }


@contextmanager
def create_dot_gersemirc(where, **kwargs):
    p = os.path.join(where, ".gersemirc")
    try:
        with open(p, "w", encoding="utf-8") as f:
            f.write(yaml.dump(kwargs))
        yield
    finally:
        os.remove(p)


@contextmanager
def create_fake_definitions(where, name):
    p = os.path.join(where, name)
    try:
        with open(os.path.join(where, name), "w", encoding="utf-8") as f:
            f.write("\n")
        yield p
    finally:
        os.remove(p)


@contextmanager
def create_configuration_files(directories, creator):
    with ExitStack() as stack:
        configuration_files = [
            stack.enter_context(create_dot_gersemirc(where=d, **creator(d)))
            for d in directories
        ]
        yield configuration_files


def assert_that_directories_differ(left, right):
    comparison = compare_directories(left, right)
    for value in comparison.values():
        if len(value) > 0:
            return
    raise AssertionError("directories have the same content")


def assert_that_directories_have_the_same_content(left, right):
    comparison = compare_directories(left, right)
    report = []
    for key, value in comparison.items():
        if len(value) > 0:
            report += [f"{key}: {value}"]

    assert len(report) == 0, "directories differences:\n" + "\n".join(report)


def test_help_just_works(app):
    assert app("--help") == success()


def test_check_on_formatted_file_should_return_zero(app, testfiles):
    assert app("--check", testfiles / "formatted_file.cmake") == success()


def test_check_on_not_formatted_file_should_return_one(app, testfiles):
    target = (testfiles / "not_formatted_file.cmake").resolve()
    assert app("--check", target) == fail(
        stdout="",
        stderr=f"""{target} would be reformatted
""",
    )


def test_diff_on_not_formatted_files_should_return_zero(app, testfiles):
    target = testfiles / "directory_with_not_formatted_files"
    assert app("--diff", target) == success(stdout=match_not(""), stderr="")


def test_check_with_diff_on_not_formatted_files_should_return_one(app, testfiles):
    target = testfiles / "directory_with_not_formatted_files"
    file1 = (target / "file1.cmake").resolve()
    file2 = (target / "file2.cmake").resolve()
    file3 = (target / "file3.cmake").resolve()

    assert app("--check", "--diff", target) == fail(
        stdout=match_not(""),
        stderr=f"""{file1} would be reformatted
{file2} would be reformatted
{file3} would be reformatted
""",
    )


def test_format_file_in_place(app, testfiles):
    f = testfiles / "not_formatted_file.cmake"
    assert app("--check", f) == fail()
    assert app("--in-place", f) == success()
    assert app("--check", f) == success()


def test_check_formatted_input_from_stdin(app):
    inp = """set(FOO BAR)
"""
    assert app("--check", "-", input=inp) == success()


def test_check_not_formatted_input_from_stdin(app):
    inp = """set(FOO BAR)"""  # missing newline at the end
    assert app("--check", "-", input=inp) == fail()


def test_format_formatted_input_from_stdin(app):
    inp = """set(FOO BAR)
"""
    assert app("-", input=inp) == success(stdout=inp, stderr="")


def test_format_not_formatted_input_from_stdin(app):
    inp = "set(FOO BAR)"  # missing newline at the end

    assert app("-", input=inp) == success(stdout=inp + "\n", stderr="")


def test_when_run_with_no_input_should_return_zero(app):
    # no args
    assert app() == success()


def test_dont_mix_stdin_and_file_input(app, testfiles):
    f = testfiles / "formatted_file.cmake"
    assert app(f, "-") == fail(stderr="Don't mix stdin with file input\n")
    assert app("-", f) == fail(stderr="Don't mix stdin with file input\n")
    assert app(f, "-", f) == fail(stderr="Don't mix stdin with file input\n")


def test_check_multiple_formatted_input_files(app, testfiles):
    files = [
        testfiles / "directory_with_formatted_files" / f
        for f in ["file1.cmake", "file2.cmake", "file3.cmake"]
    ]
    assert app("--check", *files) == success()


def test_check_multiple_not_formatted_input_files(app, testfiles):
    files = [
        testfiles / "directory_with_not_formatted_files" / f
        for f in ["file1.cmake", "file2.cmake", "file3.cmake"]
    ]
    assert app("--check", *files) == fail()


def test_check_multiple_input_files_when_some_are_not_formatted(app, testfiles):
    files = [
        testfiles / "directory_with_some_not_formatted_files" / f
        for f in [
            "formatted_file1.cmake",
            "formatted_file2.cmake",
            "formatted_file3.cmake",
            "not_formatted_file1.cmake",
            "not_formatted_file2.cmake",
            "not_formatted_file3.cmake",
        ]
    ]
    assert app("--check", *files) == fail()


def test_check_directory_with_formatted_files(app, testfiles):
    assert app("--check", testfiles / "directory_with_formatted_files") == success()


def test_check_directory_with_not_formatted_files(app, testfiles):
    assert app("--check", testfiles / "directory_with_not_formatted_files") == fail()


def test_check_directory_with_some_not_formatted_files(app, testfiles):
    assert (
        app("--check", testfiles / "directory_with_some_not_formatted_files") == fail()
    )


def test_format_in_place_multiple_formatted_files(app, testfiles):
    files = [
        testfiles / "directory_with_formatted_files" / f
        for f in ["file1.cmake", "file2.cmake", "file3.cmake"]
    ]
    assert app("--check", *files) == success()
    assert app("--in-place", *files) == success()
    assert app("--check", *files) == success()


def test_format_in_place_multiple_not_formatted_files(app, testfiles):
    files = [
        testfiles / "directory_with_not_formatted_files" / f
        for f in ["file1.cmake", "file2.cmake", "file3.cmake"]
    ]
    assert app("--check", *files) == fail()
    assert app("--in-place", *files) == success()
    assert app("--check", *files) == success()


def test_format_in_place_multiple_input_files_when_some_are_not_formatted(
    app, testfiles
):
    files = [
        testfiles / "directory_with_some_not_formatted_files" / f
        for f in [
            "formatted_file1.cmake",
            "formatted_file2.cmake",
            "formatted_file3.cmake",
            "not_formatted_file1.cmake",
            "not_formatted_file2.cmake",
            "not_formatted_file3.cmake",
        ]
    ]
    formatted, not_formatted = files[:3], files[3:]

    assert app("--check", *formatted) == success()
    assert app("--check", *not_formatted) == fail()
    assert app("--in-place", *formatted, *not_formatted) == success()
    assert app("--check", *formatted, *not_formatted) == success()


def test_format_in_place_directory_with_formatted_files(app, testfiles):
    d = testfiles / "directory_with_formatted_files"
    assert app("--check", d) == success()
    assert app("--in-place", d) == success()
    assert app("--check", d) == success()


def test_format_in_place_directory_with_not_formatted_files(app, testfiles):
    d = testfiles / "directory_with_not_formatted_files"
    assert app("--check", d) == fail()
    assert app("--in-place", d) == success()
    assert app("--check", d) == success()


def test_format_in_place_directory_with_some_not_formatted_files(app, testfiles):
    d = testfiles / "directory_with_some_not_formatted_files"
    assert app("--check", d) == fail()
    assert app("--in-place", d) == success()
    assert app("--check", d) == success()


def test_format_with_default_line_length(app):
    inp = "set(FOO long_argument__________________________________________________________)"
    assert len(inp) == 80
    assert app("-", input=inp) == success(stdout=inp + "\n", stderr="")

    inp2 = "set(FOO long_argument____________________________________________________________)"
    outp2 = """set(FOO
    long_argument____________________________________________________________
)
"""
    assert len(inp2) > 80
    assert app("-", input=inp2) == success(stdout=outp2, stderr="")


def test_format_with_non_default_line_length(app):
    line_length = 30
    inp = "set(FOO long_argument________)"
    assert len(inp) == line_length
    assert app("--line-length", line_length, "-", input=inp) == success(
        stdout=inp + "\n", stderr=""
    )

    inp2 = "set(FOO long_argument__________)"
    outp2 = """set(FOO
    long_argument__________
)
"""
    assert len(inp2) > line_length
    assert app("--line-length", line_length, "-", input=inp2) == success(
        stdout=outp2, stderr=""
    )


def test_check_project_with_custom_commands(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    assert app("--check", not_formatted, "--definitions", not_formatted) == fail()

    formatted = testfiles / "custom_project" / "formatted"
    assert app("--check", formatted, "--definitions", formatted) == success()


def test_check_project_with_custom_commands_but_without_definitions(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    assert app("--check", not_formatted) == fail()

    formatted = testfiles / "custom_project" / "only_custom_commands_not_formatted"
    assert app("--check", formatted) == success()


def test_format_project_with_custom_commands(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted"

    assert_that_directories_differ(not_formatted, formatted)
    assert app("--check", formatted, "--definitions", formatted) == success()
    assert app("--check", not_formatted, "--definitions", not_formatted) == fail()
    assert app("--in-place", not_formatted, "--definitions", not_formatted) == success()
    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_format_project_with_custom_commands_but_without_definitions(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "only_custom_commands_not_formatted"

    assert_that_directories_differ(not_formatted, formatted)
    assert app("--check", formatted) == success()
    assert app("--check", not_formatted) == fail()
    assert app("--in-place", not_formatted) == success()
    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_non_empty_stderr_when_files_are_not_formatted(app, testfiles):
    f = testfiles / "custom_project" / "not_formatted"
    assert app("--check", f) == fail(stdout="", stderr=match_not(""))


def test_empty_stderr_when_files_are_not_formatted_but_quiet_is_supplied(
    app, testfiles
):
    f = testfiles / "custom_project" / "not_formatted"
    assert app("--check", f, "--quiet") == fail(stderr="")


def test_project_with_dot_gersemirc_will_use_configuration_defined_in_file(
    app, testfiles
):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted_with_line_length_100"

    assert_that_directories_differ(not_formatted, formatted)

    def creator(dirname):
        return {"line_length": 100, "definitions": [str(dirname)]}

    with create_configuration_files([not_formatted, formatted], creator):
        assert app("--check", formatted) == success()
        assert app("--check", not_formatted) == fail()
        assert app("--in-place", not_formatted) == success()

    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_line_length_from_command_line_takes_precedence_over_configuration_file(
    app, testfiles
):
    formatted = testfiles / "custom_project" / "formatted_with_line_length_100"

    app_ = partial(app, "--check", formatted, "--definitions", formatted)

    # without configuration file
    assert app_() == fail()
    assert app_("--line-length", "100") == success()
    assert app_("--line-length", "80") == fail()

    # with configuration file
    with create_dot_gersemirc(where=formatted, line_length=100):
        assert app_() == success()
        assert app_("--line-length", "100") == success()
        assert app_("--line-length", "80") == fail()

    # and without configuration file again
    assert app_() == fail()
    assert app_("--line-length", "100") == success()
    assert app_("--line-length", "80") == fail()


def test_definitions_from_command_line(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted"

    assert_that_directories_differ(not_formatted, formatted)
    assert app("--check", formatted, "--definitions", formatted) == success()
    assert app("--check", not_formatted, "--definitions", not_formatted) == fail()
    assert app("--in-place", not_formatted, "--definitions", not_formatted) == success()
    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_definitions_from_command_line_take_precedence_over_configuration_file_with_no_definitions(
    app, testfiles
):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted"

    def empty_creator(_):
        return {}

    assert_that_directories_differ(not_formatted, formatted)

    with create_configuration_files([not_formatted, formatted], empty_creator):
        assert app("--check", formatted) == success()
        assert app("--check", not_formatted) == fail()
        assert app("--in-place", not_formatted) == success()

    assert_that_directories_differ(not_formatted, formatted)


def test_missing_definitions_in_configuration_file_but_provided_from_command_line(
    app, testfiles
):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted"

    def empty_creator(_):
        return {}

    assert_that_directories_differ(not_formatted, formatted)

    with create_configuration_files([not_formatted, formatted], empty_creator):
        assert app("--check", formatted, "--definitions", formatted) == success()
        assert app("--check", not_formatted, "--definitions", not_formatted) == fail()
        assert (
            app("--in-place", not_formatted, "--definitions", not_formatted)
            == success()
        )

    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_best_definitions_in_configuration_file_but_overridden_by_command_line(
    app, testfiles
):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted"

    assert_that_directories_differ(not_formatted, formatted)

    def creator(dirname):
        return {"definitions": [str(dirname)]}

    with create_configuration_files([not_formatted, formatted], creator):
        with ExitStack() as fake_definitions:
            definitions_in_not_formatted, definitions_in_formatted = [
                fake_definitions.enter_context(
                    create_fake_definitions(where=dirname, name="definitions.cmake")
                )
                for dirname in map(str, [not_formatted, formatted])
            ]

            assert (
                app("--check", formatted, "--definitions", definitions_in_formatted)
                == success()
            )
            assert (
                app(
                    "--check",
                    not_formatted,
                    "--definitions",
                    definitions_in_not_formatted,
                )
                == fail()
            )
            assert (
                app(
                    "--in-place",
                    not_formatted,
                    "--definitions",
                    definitions_in_not_formatted,
                )
                == success()
            )

    assert_that_directories_differ(not_formatted, formatted)


def test_use_paths_relative_to_root_as_definitions_in_configuration_file(
    app, testfiles
):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted_with_line_length_100"

    assert_that_directories_differ(not_formatted, formatted)

    def creator(_):
        return {
            "line_length": 100,
            "definitions": [
                "back_to_the_future.cmake",
                "back_to_the_future_sequels.cmake",
            ],
        }

    with create_configuration_files([not_formatted, formatted], creator):
        assert app("--check", formatted) == success()
        assert app("--check", not_formatted) == fail()
        assert app("--in-place", not_formatted) == success()

    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_use_absolute_paths_as_definitions_in_configuration_file(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"
    formatted = testfiles / "custom_project" / "formatted_with_line_length_100"

    assert_that_directories_differ(not_formatted, formatted)

    def creator(dirname):
        return {
            "line_length": 100,
            "definitions": [
                os.path.join(dirname, "back_to_the_future.cmake"),
                os.path.join(dirname, "back_to_the_future_sequels.cmake"),
            ],
        }

    with create_configuration_files([not_formatted, formatted], creator):
        assert app("--check", formatted) == success()
        assert app("--check", not_formatted) == fail()
        assert app("--in-place", not_formatted) == success()

    assert_that_directories_have_the_same_content(not_formatted, formatted)


def test_use_configuration_file_from_current_directory_when_input_is_from_stdin(
    tmpdir, app
):
    line_length = 30
    inp = "set(FOO long_argument__________)"
    outp = """set(FOO
    long_argument__________
)
"""
    assert len(inp) > line_length
    with create_dot_gersemirc(where=tmpdir, line_length=30):
        assert app("-", input=inp, cwd=tmpdir) == success(stdout=outp, stderr="")


def test_use_configuration_file_from_parent_directory_when_input_is_from_stdin(
    tmpdir, app
):
    line_length = 30
    inp = "set(FOO long_argument__________)"
    outp = """set(FOO
    long_argument__________
)
"""
    assert len(inp) > line_length

    nested_directory = tmpdir.mkdir("foo").mkdir("bar")
    app_ = partial(app, cwd=nested_directory)

    assert app_("-", input=inp) == success(stdout=match_not(outp), stderr="")
    with create_dot_gersemirc(where=tmpdir, line_length=30):
        assert app("-", input=inp) == success(stdout=outp, stderr="")


def test_formatted_files_are_stored_in_cache_on_check(app, cache, testfiles):
    d = testfiles / "custom_project" / "formatted"
    cache.assert_that_has_no_tables()

    assert app("--check", d, "--definitions", d) == success()

    cache.assert_that_has_initialized_tables()

    assert len(cache.get_files()) > 0
    assert len(cache.get_formatted()) > 0


def test_not_formatted_files_are_not_stored_in_cache_on_check(app, cache, testfiles):
    d = testfiles / "custom_project" / "not_formatted"

    cache.assert_that_has_no_tables()

    assert app("--check", d, "--definitions", d) == fail()

    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) == 0
    assert len(cache.get_formatted()) == 0


def test_not_formatted_files_are_stored_in_cache_after_formatting(
    app, cache, testfiles
):
    d = testfiles / "custom_project" / "not_formatted"

    cache.assert_that_has_no_tables()

    assert app("--in-place", d, "--definitions", d) == success()

    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) > 0
    assert len(cache.get_formatted()) > 0


def test_formatted_files_in_cache_dont_get_updated_on_subsequent_run(
    app, cache, testfiles
):
    d = testfiles / "custom_project" / "not_formatted"

    cache.assert_that_has_no_tables()

    assert app("--in-place", d, "--definitions", d) == success()
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) > 0
    formatted_after_first_run = cache.get_formatted()
    assert len(formatted_after_first_run) > 0

    assert app("--in-place", d, "--definitions", d) == success()
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) > 0
    formatted_after_second_run = cache.get_formatted()
    assert len(formatted_after_second_run) > 0

    assert formatted_after_first_run == formatted_after_second_run


def test_different_configuration_leads_to_overriding_data_stored_in_cache(
    app, cache, testfiles
):
    d = testfiles / "custom_project" / "not_formatted"

    cache.assert_that_has_no_tables()

    assert app("--in-place", d, "--definitions", d) == success()
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) > 0
    formatted_after_first_run = cache.get_formatted()
    assert len(formatted_after_first_run) > 0

    assert app("--in-place", d, "--definitions", d, "--line-length", "100") == success()
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) > 0
    formatted_after_second_run = cache.get_formatted()
    assert len(formatted_after_first_run) > 0

    only_paths_from_first_run = [path for (path, *_) in formatted_after_first_run]
    only_paths_from_second_run = [path for (path, *_) in formatted_after_second_run]

    assert only_paths_from_first_run == only_paths_from_second_run
    assert formatted_after_first_run != formatted_after_second_run


def test_no_files_are_stored_in_cache_on_diff(app, cache, testfiles):
    d = testfiles / "custom_project" / "not_formatted"

    cache.assert_that_has_no_tables()
    assert app("--diff", d, "--definitions", d) == success(stdout=match_not(""))
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) == 0
    assert len(cache.get_formatted()) == 0


def test_when_cache_cant_be_modified_it_is_ignored(app, cache, testfiles):
    d = testfiles / "custom_project" / "formatted"
    os.chmod(cache.path, S_IREAD | S_IRGRP | S_IROTH)

    cache.assert_that_has_no_tables()
    assert app("--check", d, "--definitions", d) == success(stdout="", stderr="")
    cache.assert_that_has_no_tables()


def test_when_cache_is_malformed_it_is_ignored(app, cache, testfiles):
    d = testfiles / "custom_project" / "formatted"
    with open(cache.path, "wb") as c:
        c.write("foobarbaz1231212312312312313".encode("ascii"))
        c.flush()

    with pytest.raises(sqlite3.DatabaseError):
        cache.get_files()

    assert app("--check", d, "--definitions", d) == success(stdout="", stderr="")

    with pytest.raises(sqlite3.DatabaseError):
        cache.get_files()


def test_cache_is_not_updated_when_input_is_from_stdin(app, cache):
    inp = """set(FOO BAR)
"""
    cache.assert_that_has_no_tables()
    assert app("--check", "-", input=inp) == success()

    # first use initialized tables in cache
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) == 0
    assert len(cache.get_formatted()) == 0

    assert app("--check", "-", input=inp) == success()
    cache.assert_that_has_initialized_tables()
    assert len(cache.get_files()) == 0
    assert len(cache.get_formatted()) == 0


warning_params = [
    ("default", (), SUCCESS),
    ("warnings as errors", ("--warnings-as-errors",), FAIL),
]


@pytest.mark.parametrize(
    ["warning_args", "returncode"],
    [rest for __, *rest in warning_params],
    ids=[name for name, *__ in warning_params],
)
def test_check_project_with_conflicting_command_definitions(
    app, testfiles, warning_args, returncode
):
    base = testfiles / "conflicting_definitions"
    foo1 = (base / "foo1.cmake").resolve()
    foo2 = (base / "foo2.cmake").resolve()
    assert app(
        "--check", base, "--definitions", base, *warning_args
    ) == ExpectedOutcome(
        returncode=returncode,
        stdout="",
        stderr=f"""Warning: conflicting definitions for 'foo':
(used)    {foo1}:1:10
(ignored) {foo2}:1:10
(ignored) {foo2}:5:10
(ignored) {foo2}:9:10
""",
    )

    assert app(
        "--check", "--no-quiet", base, "--definitions", base, *warning_args
    ) == ExpectedOutcome(
        returncode=returncode,
        stdout="",
        stderr=f"""Warning: conflicting definitions for 'foo':
(used)    {foo1}:1:10
(ignored) {foo2}:1:10
(ignored) {foo2}:5:10
(ignored) {foo2}:9:10
""",
    )


def test_check_project_with_conflicting_command_definitions_dont_warn_when_quiet(
    app, testfiles
):
    base = testfiles / "conflicting_definitions"
    assert app("--check", "--quiet", base, "--definitions", base) == success(
        stdout="",
        stderr="",
    )


@pytest.mark.parametrize(
    ["warning_args", "returncode"],
    [rest for __, *rest in warning_params],
    ids=[name for name, *__ in warning_params],
)
def test_format_file_with_conflicting_command_definitions(
    app, testfiles, warning_args, returncode
):
    base = testfiles / "conflicting_definitions"
    foo1 = (base / "foo1.cmake").resolve()
    foo2 = (base / "foo2.cmake").resolve()
    assert app(
        f"{base}/CMakeLists.txt",
        "--definitions",
        base,
        "--list-expansion=favour-expansion",
        *warning_args,
    ) == ExpectedOutcome(
        returncode=returncode,
        stdout="""foo(ONE
    TWO x
    THREE
        y
        z
)
""",
        stderr=f"""Warning: conflicting definitions for 'foo':
(used)    {foo1}:1:10
(ignored) {foo2}:1:10
(ignored) {foo2}:5:10
(ignored) {foo2}:9:10
""",
    )


def test_cached_result_doesnt_inhibit_printing_in_stdout_mode(app, testfiles):
    f = testfiles / "formatted_file.cmake"

    with open(f, "r", encoding="utf-8") as fh:
        file_content = fh.read()

    assert app("--check", f) == success(stdout="")
    assert app(f) == success(stdout=file_content)


@pytest.mark.parametrize(
    ["warning_args", "returncode"],
    [rest for __, *rest in warning_params],
    ids=[name for name, *__ in warning_params],
)
@pytest.mark.parametrize(
    ["args", "check_cache"],
    [
        ((), False),
        (("--check",), True),
        (("--warn-about-unknown-commands", "--check"), True),
        (("--in-place",), True),
        (("--warn-about-unknown-commands", "--in-place"), True),
        (("--diff",), False),
    ],
)
def test_warn_about_unknown_commands(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    app, testfiles, cache, args, check_cache, warning_args, returncode
):
    target = testfiles / "warn_about_unknown_commands"
    cmakelists = target / "CMakeLists.txt"
    definitions = target / "watch_movies.cmake"

    if check_cache:
        cache.assert_that_has_no_tables()

    assert app(
        *args, cmakelists, "--definitions", definitions, *warning_args
    ) == success(stderr="")

    if check_cache:
        cache.assert_that_has_initialized_tables()
        assert len(cache.get_files()) > 0
        assert len(cache.get_formatted()) > 0

    cache.clear()

    if check_cache:
        cache.assert_that_has_no_tables()

    assert app(*args, cmakelists, *warning_args) == ExpectedOutcome(
        returncode=returncode,
        stderr=f"""Warning: unknown command 'watch_nolan_movies' used at:
{str(cmakelists.resolve())}:3:1
{str(cmakelists.resolve())}:8:5

Warning: unknown command 'watch_tarantino_movies' used at:
{str(cmakelists.resolve())}:6:5
{str(cmakelists.resolve())}:11:1

""",
    )

    if check_cache:
        cache.assert_that_has_initialized_tables()
        assert len(cache.get_files()) == 0
        assert len(cache.get_formatted()) == 0


@pytest.mark.parametrize(
    ["warning_args", "returncode"],
    [rest for __, *rest in warning_params],
    ids=[name for name, *__ in warning_params],
)
@pytest.mark.parametrize(
    ["args"],
    [
        ((),),
        (("--check",),),
        (("--diff",),),
    ],
)
def test_warn_about_unknown_commands_with_stdin(
    app, testfiles, args, warning_args, returncode
):
    target = testfiles / "warn_about_unknown_commands"
    definitions = target / "watch_movies.cmake"
    with open(target / "CMakeLists.txt", "r", encoding="utf-8") as f:
        content = f.read()

    assert app(
        *args, "-", "--definitions", definitions, *warning_args, input=content
    ) == success(stderr="")

    assert app(*args, "-", *warning_args, input=content) == ExpectedOutcome(
        returncode=returncode,
        stderr="""Warning: unknown command 'watch_nolan_movies' used at:
<stdin>:3:1
<stdin>:8:5

Warning: unknown command 'watch_tarantino_movies' used at:
<stdin>:6:5
<stdin>:11:1

""",
    )


@pytest.mark.parametrize(
    ["args", "check_cache"],
    [
        ((), False),
        (("--check",), True),
        (("--in-place",), True),
        (("--diff",), False),
    ],
)
def test_dont_warn_about_unknown_commands_when_definition_arent_required(
    app, testfiles, cache, args, check_cache
):
    target = testfiles / "warn_about_unknown_commands"
    cmakelists = target / "CMakeLists.txt"

    if check_cache:
        cache.assert_that_has_no_tables()

    assert app(*args, "--no-warn-about-unknown-commands", cmakelists) == success(
        stderr=""
    )

    if check_cache:
        cache.assert_that_has_initialized_tables()
        assert len(cache.get_files()) > 0
        assert len(cache.get_formatted()) > 0

    cache.clear()

    if check_cache:
        cache.assert_that_has_no_tables()

    with create_dot_gersemirc(where=target, warn_about_unknown_commands=False):
        assert app(*args, cmakelists) == success(stderr="")

    if check_cache:
        cache.assert_that_has_initialized_tables()
        assert len(cache.get_files()) > 0
        assert len(cache.get_formatted()) > 0

    cache.clear()

    if check_cache:
        cache.assert_that_has_no_tables()

    with create_dot_gersemirc(where=target, warn_about_unknown_commands=True):
        assert app(*args, "--no-warn-about-unknown-commands", cmakelists) == success(
            stderr=""
        )

    if check_cache:
        cache.assert_that_has_initialized_tables()
        assert len(cache.get_files()) > 0
        assert len(cache.get_formatted()) > 0


def test_cache_is_disabled(app, testfiles, cache):
    target = testfiles / "custom_project" / "formatted"
    common_args = ["--check", target, "--definitions", target]

    cache.assert_that_has_no_tables()
    assert app("--no-cache", *common_args) == success()
    cache.assert_that_has_no_tables()


def test_cache_is_enabled(app, testfiles, cache):
    target = testfiles / "custom_project" / "formatted"
    common_args = ["--check", target, "--definitions", target]

    def assert_that_cache_was_used():
        cache.assert_that_has_initialized_tables()
        assert len(cache.get_files()) > 0
        assert len(cache.get_formatted()) > 0

    # enabled by default
    cache.assert_that_has_no_tables()
    assert app(*common_args) == success()
    assert_that_cache_was_used()

    cache.clear()

    # enabled explicitly
    cache.assert_that_has_no_tables()
    assert app("--cache", *common_args) == success()
    assert_that_cache_was_used()


def test_definition_path_doesnt_exist(app, testfiles):
    project = testfiles / "custom_project"
    sources = project / "formatted"
    definitions = project / "formatted"
    common_args = ["--check", sources, "--definitions"]

    assert app(*common_args, definitions) == success()
    assert app(*common_args, project / "this_path_doesnt_exist") == fail()
    assert app(*common_args, definitions / ".." / "formatted") == success()
    assert app(*common_args, definitions / ".." / ".." / "formatted") == fail()

    definitions_as_files = [
        definitions / "back_to_the_future.cmake",
        definitions / "back_to_the_future_sequels.cmake",
    ]
    assert app(*common_args, *definitions_as_files) == success()

    definitions_as_files_with_path_that_doesnt_exist = [
        definitions / "back_to_the_future.cmake",
        definitions / "back_to_the_future_sequels.cmake",
        definitions / "back_to_the_future_four.cmake",
    ]
    assert (
        app(*common_args, *definitions_as_files_with_path_that_doesnt_exist) == fail()
    )


def test_source_path_doesnt_exist(app, testfiles):
    project = testfiles / "custom_project"
    sources = project / "formatted"
    definitions = project / "formatted"
    common_args = ["--check", sources, "--definitions"]

    assert app(*common_args, sources) == success()
    assert app(*common_args, Path(project) / "this_path_doesnt_exist") == fail()
    assert app(*common_args, sources / ".." / "formatted") == success()
    assert app(*common_args, definitions / ".." / ".." / "formatted") == fail()

    sources_as_files = [
        sources / "CMakeLists.txt",
        sources / "subdirectory_one" / "CMakeLists.txt",
        sources / "subdirectory_two" / "CMakeLists.txt",
    ]
    assert app(*common_args, *sources_as_files) == success()

    sources_as_files_with_path_that_doesnt_exist = [
        sources / "CMakeLists.txt",
        sources / "this_subdirectory_doesnt_exist" / "CMakeLists.txt",
        sources / "subdirectory_one" / "CMakeLists.txt",
        sources / "subdirectory_two" / "CMakeLists.txt",
    ]
    assert app(*common_args, *sources_as_files_with_path_that_doesnt_exist) == fail()


def test_just_works_with_not_supported_options_in_configuration_file(app, testfiles):
    f = testfiles / "directory_with_formatted_files"
    dotfile = (testfiles / ".gersemirc").resolve()
    with create_dot_gersemirc(where=testfiles, command_definitions=100):
        assert app("--check", f) == success(
            stdout="",
            stderr=f"{dotfile}: these options are not supported: command_definitions\n",
        )

    with create_dot_gersemirc(
        where=testfiles,
        line_length=80,
        kambei=1,
        katsushiro=2,
        list_expansion="favour-inlining",
        gorobei="3",
        heihachi=True,
    ):
        assert app("--check", f) == success(
            stdout="",
            # pylint: disable=line-too-long
            stderr=f"{dotfile}: these options are not supported: gorobei, heihachi, kambei, katsushiro\n",
        )

    with create_dot_gersemirc(
        where=testfiles,
        cache=False,
        color=True,
        workers=8,
        quiet=False,
        line_length=80,
        kyuzo=123,
        shichiroji="foo",
        kikuchiyo=["foo", "bar"],
    ):
        assert app("--check", f) == success(
            stdout="",
            # pylint: disable=line-too-long
            stderr=f"""{dotfile}: these options are supported only through command line: cache, color, quiet, workers
{dotfile}: these options are not supported: kikuchiyo, kyuzo, shichiroji
""",
        )


def test_when_quiet_not_supported_options_in_configuration_file_are_not_printed(
    app, testfiles
):
    f = testfiles / "formatted_file.cmake"
    with create_dot_gersemirc(where=testfiles, command_definitions=100):
        assert app("--quiet", "--check", f) == success(stdout="", stderr="")

    with create_dot_gersemirc(
        where=testfiles,
        line_length=80,
        kambei=1,
        katsushiro=2,
        list_expansion="favour-inlining",
        gorobei="3",
        heihachi=True,
    ):
        assert app("--quiet", "--check", f) == success(stdout="", stderr="")

    with create_dot_gersemirc(
        where=testfiles,
        cache=False,
        color=True,
        workers=8,
        quiet=False,
        line_length=80,
        kyuzo=123,
        shichiroji="foo",
        kikuchiyo=["foo", "bar"],
    ):
        assert app("--quiet", "--check", f) == success(stdout="", stderr="")


def test_each_file_uses_closest_configuration_file(app, testfiles):
    d = testfiles / "closest_configuration_file"
    assert app("--check", d) == success(stdout="", stderr="")


def test_each_file_uses_closest_configuration_file_with_some_not_supported_options(
    app, testfiles
):
    d = testfiles / "closest_configuration_file"
    d40 = d
    d60 = d / "60_different_config_than_root"
    d80 = (
        d
        / "mixed_config_80_subdirectory_40_files_not_in_subdirectory"
        / "80_subdirectory"
    )

    d40_dotfile = (d40 / ".gersemirc").resolve()
    d60_dotfile = (d60 / ".gersemirc").resolve()
    d80_dotfile = (d80 / ".gersemirc").resolve()

    with ExitStack() as config_files:

        def cf(**kwargs):
            config_files.enter_context(create_dot_gersemirc(**kwargs))

        cf(where=d80, line_length=80, cache=False, gorobei="3")
        cf(where=d40, line_length=40, command_definitions=100)
        cf(where=d60, line_length=60, kambei=100, katsushiro="abc")

        assert app("--check", d) == success(
            stdout="",
            stderr=f"""{d40_dotfile}: these options are not supported: command_definitions
{d60_dotfile}: these options are not supported: kambei, katsushiro
{d80_dotfile}: these options are supported only through command line: cache
{d80_dotfile}: these options are not supported: gorobei
""",
        )


def test_disable_formatting(app, testfiles):
    not_formatted = testfiles / "custom_project" / "not_formatted"

    assert app("--check", not_formatted) == fail()

    with create_dot_gersemirc(where=not_formatted, disable_formatting=True):
        assert app("--check", not_formatted) == success()
        assert app("--enable-formatting", "--check", not_formatted) == fail()
        assert app("--disable-formatting", "--check", not_formatted) == success()

    with create_dot_gersemirc(where=not_formatted, disable_formatting=False):
        assert app("--check", not_formatted) == fail()
        assert app("--enable-formatting", "--check", not_formatted) == fail()
        assert app("--disable-formatting", "--check", not_formatted) == success()


def test_disable_formatting_with_multiple_configuration_files_at_play(app, testfiles):
    d = testfiles / "closest_configuration_file"
    d40 = d
    d60 = d / "60_different_config_than_root"
    d80 = (
        d
        / "mixed_config_80_subdirectory_40_files_not_in_subdirectory"
        / "80_subdirectory"
    )

    def cf(**kwargs):
        config_files.enter_context(create_dot_gersemirc(**kwargs))

    with ExitStack() as config_files:
        cf(where=d40, line_length=60, disable_formatting=True)
        cf(where=d60, line_length=80, disable_formatting=True)
        cf(where=d80, line_length=100, disable_formatting=True)

        assert app("--check", d40) == success()
        assert app("--check", d60) == success()
        assert app("--check", d80) == success()
        assert app("--check", d) == success()

    with ExitStack() as config_files:
        cf(where=d40, line_length=70, disable_formatting=False)
        cf(where=d60, line_length=60, disable_formatting=False)
        cf(where=d80, line_length=80, disable_formatting=False)

        assert app("--check", d40) == fail()
        assert app("--check", d60) == success()
        assert app("--check", d80) == success()
        assert app("--check", d) == fail()

    with ExitStack() as config_files:
        cf(where=d80, line_length=100, disable_formatting=False)
        cf(where=d40, line_length=60, disable_formatting=False)
        cf(where=d60, line_length=80, disable_formatting=False)

        assert app("--check", d40) == fail()
        assert app("--check", d60) == fail()
        assert app("--check", d80) == fail()
        assert app("--check", d) == fail()


def test_config_parameter(app, testfiles):
    base = testfiles
    target = base / "directory_with_formatted_files"
    configuration_file_elsewhere = base / "directory_with_not_formatted_files"
    configuration_file = configuration_file_elsewhere / ".gersemirc"
    config = ["--config", configuration_file_elsewhere / ".gersemirc"]

    assert app("--check", target) == success(stderr="")
    assert app("-l", 100, "--check", target) == fail()

    with create_dot_gersemirc(where=configuration_file_elsewhere, line_length=100):
        assert app("--check", target) == success(stderr="")
        assert app("-l", 100, "--check", target) == fail()

        assert app(*config, "--check", target) == fail()
        assert app(*config, "-l", 80, "--check", target) == success(stderr="")

    with create_dot_gersemirc(
        where=configuration_file_elsewhere, line_length=100, color=True, kambei=314
    ):
        assert app("--check", target) == success(stderr="")
        assert app("-l", 100, "--check", target) == fail()

        assert app(*config, "--check", target) == fail()
        assert app(*config, "-l", 80, "--check", target) == success(
            # pylint: disable=line-too-long
            stderr=f"""{configuration_file}: these options are supported only through command line: color
{configuration_file}: these options are not supported: kambei
"""
        )



from dataclasses import dataclass
from typing import Sequence, Tuple, Union


Position = Tuple[int, int]


@dataclass
class UnknownCommandWarning:
    command_name: str
    positions: Sequence[Position]

    def get_message(self, filepath: str) -> str:
        return "\n".join(
            [
                f"Warning: unknown command '{self.command_name}' used at:",
                *(f"{filepath}:{line}:{column}" for line, column in self.positions),
                "",
            ]
        )


class WrongFormattingWarning:
    def get_message(self, filepath: str) -> str:
        return f"{filepath} would be reformatted"


FormatterWarning = Union[UnknownCommandWarning, WrongFormattingWarning]
FormatterWarnings = Sequence[FormatterWarning]



from collections.abc import Sized
from typing import Dict, Iterator, Iterable, List, Optional, Sequence, Tuple
from lark import Tree
from gersemi.ast_helpers import (
    is_comment,
    is_multi_value_argument,
    is_one_of_keywords,
    is_one_value_argument,
    is_option_argument,
    is_positional_arguments,
    is_section,
    option_argument,
    one_value_argument,
    multi_value_argument,
    positional_arguments,
)
from gersemi.base_command_invocation_dumper import BaseCommandInvocationDumper
from gersemi.keywords import KeywordMatcher
from gersemi.keyword_kind import KeywordKind, kind_to_formatter
from gersemi.types import Nodes
from gersemi.utils import pop_all


def is_non_empty(sequence: Sized) -> bool:
    return len(sequence) > 0


def is_non_empty_group(group: Sized) -> bool:
    if isinstance(group, Tree):
        return is_non_empty(group.children)

    return is_non_empty(group)


class PositionalArguments(list):
    pass


class KeywordSplitter:
    def __init__(self, options, one_value_keywords, multi_value_keywords):
        self.is_one_of_options = is_one_of_keywords(options)
        self.is_one_of_one_value_keywords = is_one_of_keywords(one_value_keywords)
        self.is_one_of_multi_value_keywords = is_one_of_keywords(multi_value_keywords)
        self.groups: List[Nodes] = []
        self.accumulator: Nodes = []
        self.comment_accumulator: Nodes = []

    def _flush_accumulators(self):
        if isinstance(self.accumulator, PositionalArguments):
            argument_group = positional_arguments(pop_all(self.accumulator))
        else:
            argument_group = multi_value_argument(pop_all(self.accumulator))
        self.groups += [argument_group, *pop_all(self.comment_accumulator)]

    def _append_option_group(self, argument):
        self._flush_accumulators()
        self.groups.append(option_argument([argument]))

    def _append_one_value_group(self, argument, iterator):
        self._flush_accumulators()
        group = [argument]
        for n in iterator:
            group.append(n)
            if not is_comment(n):
                break
        self.groups.append(one_value_argument(group))

    def split(self, arguments: Nodes) -> Iterator[Nodes]:
        iterator = iter(arguments)
        for argument in iterator:
            if is_comment(argument):
                self.comment_accumulator += [argument]
            elif self.is_one_of_options(argument):
                self._append_option_group(argument)
            elif self.is_one_of_one_value_keywords(argument):
                self._append_one_value_group(argument, iterator)
            elif self.is_one_of_multi_value_keywords(argument):
                self._flush_accumulators()
                self.accumulator = [argument]
            elif is_non_empty(self.accumulator):
                self.accumulator += [*pop_all(self.comment_accumulator), argument]
            else:
                if not isinstance(self.accumulator, PositionalArguments):
                    self.accumulator = PositionalArguments()
                self.accumulator += [*pop_all(self.comment_accumulator), argument]

        self._flush_accumulators()
        return filter(is_non_empty_group, self.groups)


class ArgumentAwareCommandInvocationDumper(BaseCommandInvocationDumper):
    _inhibit_favour_expansion: bool = False
    _keyword_formatters: Dict[str, str] = {}
    _canonical_name: Optional[str] = None

    front_positional_arguments: Sequence[str] = []
    back_positional_arguments: Sequence[str] = []
    options: Iterable[KeywordMatcher] = []
    one_value_keywords: Iterable[KeywordMatcher] = []
    multi_value_keywords: Iterable[KeywordMatcher] = []
    keyword_kinds: Dict[str, KeywordKind] = {}

    def _default_format_values(self, values) -> str:
        return "\n".join(map(self.visit, values))

    def positional_arguments(self, tree) -> str:
        return "\n".join(map(self.visit, tree.children))

    def _format_non_option(self, tree):
        result = self._try_to_format_into_single_line(tree.children)
        if result is not None:
            return result

        keyword, *values = tree.children
        keyword_as_value = keyword.children[0] if len(keyword.children) > 0 else None

        can_be_inlined = (not self.favour_expansion) or (
            self.favour_expansion
            and keyword is not None
            and keyword_as_value not in self.multi_value_keywords
        )
        if can_be_inlined:
            with self.select_inlining_strategy():
                result = self._try_to_format_into_single_line(tree.children)
                if result is not None:
                    return result

        begin = self.visit(keyword)
        if len(values) == 0:
            return begin

        formatter_kind = kind_to_formatter(
            self.keyword_kinds.get(keyword_as_value, None)
        )
        if formatter_kind is None:
            formatter_kind = self._keyword_formatters.get(
                keyword_as_value, "_default_format_values"
            )

        with self.indented():
            formatter = getattr(self, formatter_kind)
            formatted_values = formatter(values)
        return f"{begin}\n{formatted_values}"

    def _split_positional_arguments(self, arguments: Nodes, known_positional_arguments):
        last_index = min(len(arguments), len(known_positional_arguments))
        result = []
        for i in range(last_index):
            result.append(positional_arguments([arguments[i]]))

        rest = arguments[last_index:]
        if len(rest) > 0:
            result.append(positional_arguments(rest))
        return result

    def _separate_front(self, arguments: Nodes) -> Tuple[List[Nodes], Nodes]:
        is_keyword = is_one_of_keywords(
            list(self.options)
            + list(self.one_value_keywords)
            + list(self.multi_value_keywords)
        )
        for index, argument in enumerate(arguments):
            if is_keyword(argument):
                pivot = index
                break
        else:
            return (
                self._split_positional_arguments(
                    arguments, self.front_positional_arguments
                ),
                [],
            )
        return (
            self._split_positional_arguments(
                arguments[:pivot], self.front_positional_arguments
            ),
            arguments[pivot:],
        )

    def _split_by_keywords(self, arguments: Nodes) -> Iterator[Nodes]:
        splitter = KeywordSplitter(
            self.options,
            self.one_value_keywords,
            self.multi_value_keywords,
        )
        return splitter.split(arguments)

    def _split_arguments(self, arguments):
        if len(self.back_positional_arguments) > 0:
            arguments, back = (
                arguments[: -len(self.back_positional_arguments)],
                [
                    positional_arguments(
                        arguments[-len(self.back_positional_arguments) :]
                    )
                ],
            )
        else:
            back = []
        front, tail = self._separate_front(arguments)
        keyworded_arguments = self._split_by_keywords(tail)

        return [*front, *keyworded_arguments, *back]

    def group_size(self, group):
        if is_positional_arguments(group):
            return len(group.children)
        if is_option_argument(group):
            return 0
        if is_one_value_argument(group):
            return len(group.children) - 1
        if is_multi_value_argument(group):
            return len(group.children) - 1
        if is_section(group):
            section_size = len(group.children) - 1
            subarguments_size = max(map(self.group_size, group.children))
            return max(section_size, subarguments_size)

        return 0

    def option_argument(self, tree):
        return self.visit(tree.children[0])

    def one_value_argument(self, tree):
        return self._format_non_option(tree)

    def multi_value_argument(self, tree):
        return self._format_non_option(tree)

    def arguments(self, tree):
        groups = self._split_arguments(tree.children)
        return "\n".join(map(self.visit, filter(None, groups)))

    def format_command_name(self, identifier):
        if self._canonical_name is None:
            return super().format_command_name(identifier)

        if self._canonical_name.lower() != identifier.lower():
            raise RuntimeError

        return self._canonical_name



from enum import Enum
from typing import Optional, Union


class KeywordKind(Enum):
    CommandLine = "command_line"
    Pairs = "pairs"


def get_kind(kind: Union[None, str, KeywordKind]) -> Optional[KeywordKind]:
    if isinstance(kind, KeywordKind):
        return kind

    return KeywordKind(kind) if kind in [e.value for e in KeywordKind] else None


def kind_to_formatter(kind: Union[None, str, KeywordKind]) -> Optional[str]:
    proper_kind = get_kind(kind)
    if proper_kind is None:
        return None

    return {
        KeywordKind.CommandLine: "_format_command_line",
        KeywordKind.Pairs: "_format_keyword_with_pairs",
    }.get(proper_kind, None)



from dataclasses import dataclass
from pathlib import Path
from gersemi.warnings import FormatterWarnings


@dataclass
class TaskResult:
    path: Path
    return_code: int
    to_stdout: str = ""
    to_stderr: str = ""
    warnings: FormatterWarnings = tuple()



from collections.abc import Mapping, Collection
from contextlib import contextmanager
from functools import lru_cache
import importlib
import string
from gersemi.builtin_commands import preprocess_definitions


class VerificationFailure(Exception):
    pass


PROPERTIES_WITH_KEYWORDS = {
    "options",
    "one_value_keywords",
    "multi_value_keywords",
}
PROPERTIES_WITH_POSITIONAL_ARGUMENTS = {
    "front_positional_arguments",
    "back_positional_arguments",
}

IDENTIFIER_START = string.ascii_letters + "_"


def is_identifier(identifier):
    return (len(identifier) > 0) and (identifier[0] in IDENTIFIER_START)


class Verifier:
    def __init__(self, extension):
        self.base = f"gersemi_{extension}.command_definitions"

    @contextmanager
    def element(self, name):
        old = self.base
        try:
            self.base = f"{self.base}[{repr(name)}]"
            yield self
        finally:
            self.base = old

    def fail(self, reason):
        raise VerificationFailure(f"{self.base}: {reason}")

    def verify_collection_of_arguments(self, collection):
        if not isinstance(collection, Collection):
            self.fail(f"is not a collection of strings ({repr(collection)})")

        for index, item in enumerate(collection):
            if not isinstance(item, str):
                with self.element(index):
                    self.fail(f"argument ({repr(item)}) has to be a string")

    def verify_identifier(self, thing, keyword):
        if not isinstance(keyword, str):
            self.fail(f"{thing} ({repr(keyword)}) has to be a string")

        if not is_identifier(keyword):
            self.fail(
                f"{thing} ({repr(keyword)}) has to start with a letter or underscore"
            )

    def verify_collection_of_keywords(self, collection):
        if not isinstance(collection, Collection):
            self.fail(f"is not a collection of strings ({repr(collection)})")

        for keyword in collection:
            self.verify_identifier("keyword", keyword)

    def verify_is_mapping(self, thing):
        if not isinstance(thing, Mapping):
            self.fail("is not a mapping")

    def verify_section(self, section):
        self.verify_is_mapping(section)

        for p in PROPERTIES_WITH_POSITIONAL_ARGUMENTS:
            with self.element(p):
                self.verify_collection_of_arguments(section.get(p, []))

        for p in PROPERTIES_WITH_KEYWORDS:
            with self.element(p):
                self.verify_collection_of_keywords(section.get(p, []))

        subsections = section.get("sections", {})
        with self.element("sections"):
            self.verify_is_mapping(subsections)

            for keyword, subsection in subsections.items():
                self.verify_identifier("keyword", keyword)

                with self.element(keyword):
                    self.verify_section(subsection)

    def verify_signatures(self, signatures):
        self.verify_is_mapping(signatures)

        for signature_name, signature in signatures.items():
            self.verify_identifier("signature", signature_name)
            with self.element(signature_name):
                self.verify_section(signature)

    def verify_command(self, name, definition):
        self.verify_identifier("command name", name)

        with self.element(name):
            self.verify_is_mapping(definition)

            if "signatures" in definition:
                with self.element("signatures"):
                    self.verify_signatures(definition["signatures"])
            else:
                self.verify_section(definition)

    def __call__(self, thing):
        self.verify_is_mapping(thing)

        for name, definition in thing.items():
            self.verify_command(name, definition)


def verify(extension, thing):
    verifier = Verifier(extension)
    verifier(thing)


@lru_cache(maxsize=None)
def load_definitions_from_extension(extension):
    try:
        module = importlib.import_module(f"gersemi_{extension}")
    except ModuleNotFoundError:
        return None, f"Missing extension {extension}"

    try:
        command_definitions = module.command_definitions
    except AttributeError:
        return None, f"Extension {extension} doesn't implement command_definitions"

    try:
        verify(extension, command_definitions)
    except VerificationFailure as failure:
        return (
            None,
            f"""Verification failed for extension {extension}:
{failure}""",
        )

    return preprocess_definitions(command_definitions), None


def load_definitions_from_extensions(extensions):
    result = {}
    errors = []

    for extension in extensions:
        try:
            definitions, error = load_definitions_from_extension(extension)
            if error is not None:
                errors.append(error)
            else:
                result.update(definitions)
        except Exception as e:  # pylint: disable=broad-exception-caught
            errors.append(str(e))

    if errors:
        raise Exception(  # pylint: disable=broad-exception-raised
            "\n".join(sorted(errors))
        )

    return result
